# Registry of data files with their SHA256 hashes
# To generate the SHA256 hash, use the command:
# openssl sha256 filename

registry = {
    # Concentration Targets
    "concentration_targets/cn_sample_2.tiff": "a4c37bd36042370567297a6802a623586dc86405f11b14e32c15232849aa06be",
    "concentration_targets/cn_sample_3.tiff": "aac12c12735f2de5563ab5177361502d2f91386b542dfea949163601392fc54a",
    "concentration_targets/cn_sample_1.tiff": "8dae075be3cb52d849b1b4aaf801036ccc9103ee9ccba9c11aeb055cb2fcdab7",
    "concentration_targets/cn_sample_4.tiff": "784e376feafa0f5c8624fd437bb70d416170e605b00213d849975073a405d057",

    # Depth Targets
    "depth_targets/depth_sample_1.tiff": "da8a4e6d9f4d7e13048462b06a2efb0f5da10ea4d4d36be2ee07cd3b73c8dbd1",
    "depth_targets/depth_sample_2.tiff": "3d5fefd4d4335fd682d253bd8a33fc2cbe9dfd09660051e068bc217816ba5c3d",

    # Resolution Targets
    "resolution_targets/resolution_target_cropped.tiff": "6eecae1044587f2a2598ea97b017cab7c20935fee0d150068dd508523c01e581",
    "resolution_targets/res_sample_1.tiff": "a0f967202b36ebe732058fb6ab23b952773bb025a3e958752cc232c7a54d8b2d",

    # Depth Resolution Targets
    "depth_resolution_targets/dr_sample1.tiff": "58790cc0f3b17e008260337b3ae24457f61fd41d77baaae16570516147439990",
    "depth_resolution_targets/dr_sample2.tiff": "08ff21d678db1b0f29a3ed09fe4e3062d20a6541d32dccfb737e748aa586f49c",

    # Lung Test Image
    "lung_test_image/lung_test_image.tiff": "fb8bc5a98428cf0acb19693b241805f07c4cd888186820f60fb485d686b3a0fe",

    # Lung Reference Source
    "lung_reference_source/Phantom image.tiff": "26ec41433735675274476a065792c99abdd34ddb58c6968a08077575b7fee5b0",
    "lung_reference_source/Inclusions image.tiff": "23c2fee447d5cde3ea965b324fd9afb6be4b68c138fba1af4afb65402f84a458",

    # RUD Targets - Example 1
    "rud_targets/example_1/rud_image1.tiff": "4cde7ef6639898685133378a87b4171e5ead1d006519df2279507cbacce1bbdb",
    "rud_targets/example_1/rud_image2.tiff": "45e6b96e1a638c2ad4a26b4cf7c98ced5d36dc6e63ba5a01f6e9907dde936281",
    "rud_targets/example_1/rud_image3.tiff": "fd1b75ce2cb2e20bfb5c2cd8cf875c4697f92ec3f54f2e0b816b1f705f9723e2",
    "rud_targets/example_1/rud_image4.tiff": "0d168e244b93c3cc4cf0c4a0f272752d6c79322fc2f970cc2e924d70e90ef346",

    # RUD Targets - Example 2
    "rud_targets/example_2/rud_image1.tiff": "382728a8810f4e50e29de53263a1334d9b323e784957703a7be759d18345193e",
    "rud_targets/example_2/rud_image3.tiff": "7fd4ceb9ef9620264593d6937e08f1d2b1db448ae9f23ae5db818bc2d3d1833c",
    "rud_targets/example_2/rud_image2.tiff": "308da7d7d76cadc00063f2bb20fda61e1b9f9e0589ae7523f518ecc58c2b9ac9",
    "rud_targets/example_2/rud_image5.tiff": "f0ef2923a4829ef08b46d904b6a6dea272069cd693595ef41f58663fc3b5c174",
    "rud_targets/example_2/rud_image4.tiff": "52a7ecae992d4ce9ce932d22527f422f029ca955935fca7ce8e10ccaff1ced1d",

    # RUD Targets - Example 3
    "rud_targets/example_3/rud_center_fov.tiff": "984cc3cab573c06c1ab3c0b4b2db5e14038cab18c3feacdf98598ca9370f7b19",

    # RUD Targets - Example 4
    "rud_targets/example_4/rud_image.TIF": "5ed18c9ad867348f90597c158b5b4cbf1c20b3750d149cce40e2c25656b8fe71",
}
